## ------------------ ##

# Derive shipping routes
# Adapting to a Foggy Future along Trans-Arctic Shipping Routes
# 6 December 2022



## ------------------ ##

##### install packages  #####

# The raster package (Robert J.Hijmans 2019)
# More details:https://rspatial.org/raster/RasterPackage.pdf
install.packages("raster")

# R Package gdistance: Distances and Routes on Geographical Grids (Jacob van Etten 2017)
# More details: https://mran.microsoft.com/snapshot/2017-02-04/web/packages/gdistance/vignettes/gdistance1.pdf
install.packages("gdistance")

# A* algorithm implemented in R. (Michael Chow 2018)
remotes::install_github('machow/astar-r')

# The RMatlab Package (Liu and Davis 2007)
# More details: https://www.omegahat.net/RMatlab/outline.pdf
install.packages('R.matlab')

##### load packages #####
library("gdistance")
library("astar")
library("R.matlab")
library("raster")
library("data.table")
library("dplyr")
library("rlist")

##### function #####

# Get nodes for A* algorithm
get_nodes <- function(target_data) {
  num<-dim(target_data)[1]
  nodes <- list()
  id<-as.character( 1:num)
  for(i in 1:num){
    assign(paste0('p',i),unlist(1/target_data[i,1:num]))
    a<-get(paste0('p',i))
    names(a)<-id
    a<-lapply(list(a), function(x) x[!is.na(as.numeric(x)) & x != Inf])
    nodes[i]<-a
    if(i%%1000==0){print(i)}
  }
  return(nodes)
}

# Caculate the routes and derive a dataframe (position speed time distance)
get_path.dataframe <- function(start,goal,r,nodes,hd,CorrMatrix) {
  
  # A* algorithm
  start<-as.character(cellFromXY(r_newproj,start))
  goal<-as.character(cellFromXY(r_newproj,goal))
  
  neighbors <- function(node) names(nodes[[node]])
  cost_estimate <- function(node, goal) 1
  fscore = cost_estimate(start, goal)
  edge_distance <- function(src, dst) nodes[[src]][dst] 
  is_goal_reached <- function(node, goal) identical(node, goal) 
  
  path <- astar(start, goal, cost_estimate, edge_distance, neighbors, is_goal_reached)
  
  # dataframe
  path<-as.numeric(  unlist(path)  ) 
  dataframe <- matrix(nr=length(path),nc=6)
  dataframe<-data.frame(dataframe)
  names(dataframe)<-c("x","y",'speed',"CorrMatrix","time","distance")
  
  hd_raster<-raster(hd)
  CorrMatrix_raster<-raster(CorrMatrix)
  
  
  for(i in 1:length(path)){
    cell<-xyFromCell(r_newproj, path[i])
    dataframe[i,]<- c(cell[1],cell[2],r_newproj[path[i]]*1852,CorrMatrix_raster[path[i]],0,0)
  }
  
  
  dataframe$time<-1/dataframe$CorrMatrix
  dataframe$distance<-dataframe$speed*dataframe$time
  return(dataframe)
}

# Plot routes on map
get_routes<-function(dataframe,color) {
  coords<-cbind(dataframe$x,dataframe$y)
  lines(coords, col=color)
  return(dataframe)
}

##### Set starting and ending points #####
#+proj=aeqd +lat_0=80 +lon_0=10  +x_0=4321000 +y_0=3210000 +datum=WGS84 +units=m +no_defs
A<-c(4281387.5546441,7045886.58048823)#Bering strait
B<-c(3926347.40696818,150112.53967471)#Rotterdam
C<-c(199664.30455789,1963820.4914725)#St'Jhon

##### Select Data and Vessel #####
rcp     =  "45"      # 85 or 45
vessel  =  "Type A"    # Ship Category
year=seq(2006,2096,10) # Time
j=ifelse(vessel=="Type A",10000,1)

# Route-I
pathread=paste("/speed_raster/",vessel,"/rcp",rcp,sep="")
pathsave=paste("/path_nofog/path",rcp,'/',vessel,"/",sep="")
# Route-F
pathread=paste("/speed_raster_fog/",vessel,"/rcp",rcp,sep="")
pathsave=paste("/path_fog/path",rcp,'/',vessel,"/",sep="")


##### Route visualization #####
# read speed raster [nlat nlon ntime]
pathname<-file.path(pathread,'speed_raster.mat')
data<-readMat(pathname)
dim_speed<-dim(data[["speed"]])
mydata<-array(data[["speed"]],dim=c(dim_speed[1],dim_speed[2],dim_speed[3]))
speed<-as.matrix(mydata[,,1])
# Transfer to a newproj
r<-raster(ncol=dim_speed[2], nrow=dim_speed[1], xmn=-180, xmx=180, ymn=-90, ymx=90)
r[]<-speed
newproj= "+proj=aeqd +lat_0=80 +lon_0=10  +x_0=4321000 +y_0=3210000 +datum=WGS84 +units=m +no_defs"
r_newproj <- projectRaster(r, crs = newproj, method = "ngb") 
par(mar=c(4, 6 ,4, 4) + 0.1,mai=c(0.82,1.22,0.82,0.82))
image(r_newproj, main="", col=rev(terrain.colors(255)), xlab="x coordinate (m)", ylab="y coordinate (m)")

##### Route caculation #####
for(l in 1:dim_speed[3]){
  data<-readMat(pathname)
  mydata<-array(data[["speed"]],dim=c(dim_speed[1],dim_speed[2],dim_speed[3]))
  speed<-as.matrix(mydata[,,l])
  rm(mydata,data)
  r<-raster(ncol=dim_speed[2], nrow=dim_speed[1], xmn=-180, xmx=180, ymn=-90, ymx=90)
  r[]<-speed
  newproj= "+proj=aeqd +lat_0=80 +lon_0=10  +x_0=4321000 +y_0=3210000 +datum=WGS84 +units=m +no_defs"
  r_newproj <- projectRaster(r, crs = newproj, method = "ngb")
  
  # get routes
  speed_km_min<- function(x) max(x) 
  hd <- transition(r_newproj, speed_km_min,8,symm=F)
  hd<-hd*1852# Unit:km/h
  CorrMatrix<-geoCorrection(hd, type="c", multpl=FALSE, scl=FALSE)
  
  time<- transitionMatrix(CorrMatrix)
  dim(time)
  
  gc()
  memory.limit(5000000)
  
  t1=proc.time()
  
  nodes<-get_nodes(target_data=time)
  
  t2=proc.time()
  t=t2-t1
  print(paste0('The',l,'time---Spend',t[3][[1]]/3600,'h'))
  
  nodes<-lapply(nodes, function(x) x[!is.na(as.numeric(x)) & x != Inf])
  names(nodes)<-as.character( 1:dim(time)[1])
  
  # get dataframe and print total sailing time and distance
  dataframe_BtoA<-get_path.dataframe(start=B,goal=A,nodes=nodes,hd=hd,CorrMatrix=CorrMatrix)
  dataframe_BtoA<-transform(dataframe_BtoA,ID=l*j,year=year[l],rcp=rcp,vessel=vessel)
  print(paste('NSR',sum(dataframe_BtoA$time)/24,'d',sum(dataframe_BtoA$distance)/1852,'n mile'))
  
  dataframe_CtoA<-get_path.dataframe(start=C,goal=A,nodes=nodes,hd=hd,CorrMatrix=CorrMatrix)
  dataframe_CtoA<-transform(dataframe_CtoA,ID=(l*100)*j,year=year[l],rcp=rcp,vessel=vessel)
  print(paste('NWP',sum(dataframe_CtoA$time)/24,'d',sum(dataframe_CtoA$distance)/1852,'n mile'))
  
  # plot routes
  dataframe_BtoA<-get_routes(dataframe_BtoA,"red")
  dataframe_CtoA<-get_routes(dataframe_CtoA,"blue")
  
  # save data
  filename1=paste(pathsave,"B_R20",l-1,"6.csv",sep="")
  write.csv(dataframe_BtoA,file =filename1 ,row.names = TRUE)
  filename2=paste(pathsave,"B_S20",l-1,"6.csv",sep="")
  write.csv(dataframe_CtoA,file =filename2 ,row.names = TRUE)
  
}